# Copyright 2025 © BeeAI a Series of LF Projects, LLC
# SPDX-License-Identifier: Apache-2.0

import asyncio
import uuid
from collections.abc import AsyncIterator
from datetime import datetime, timezone
from enum import Enum
from typing import Any, Literal, Optional, Union

from pydantic import AnyUrl, BaseModel, ConfigDict, Field
from pydantic.json_schema import SkipJsonSchema

from acp_sdk.models.common import AnyModel
from acp_sdk.models.errors import ACPError, Error
from acp_sdk.models.platform import PlatformUIAnnotation
from acp_sdk.models.types import AgentName, ResourceUrl, RunId, SessionId
from acp_sdk.shared import ResourceLoader, ResourceStore


class Author(BaseModel):
    name: str
    email: str | None = None
    url: AnyUrl | None = None


class Contributor(BaseModel):
    name: str
    email: str | None = None
    url: AnyUrl | None = None


class LinkType(str, Enum):
    SOURCE_CODE = "source-code"
    CONTAINER_IMAGE = "container-image"
    HOMEPAGE = "homepage"
    DOCUMENTATION = "documentation"


class Link(BaseModel):
    type: LinkType
    url: AnyUrl


class DependencyType(str, Enum):
    AGENT = "agent"
    TOOL = "tool"
    MODEL = "model"


class Dependency(BaseModel):
    type: DependencyType
    name: str


class Capability(BaseModel):
    name: str
    description: str


class Annotations(BaseModel):
    beeai_ui: PlatformUIAnnotation | None = None
    model_config = ConfigDict(extra="allow")


class Metadata(BaseModel):
    annotations: Annotations | None = None
    documentation: str | None = None
    license: str | None = None
    programming_language: str | None = None
    natural_languages: list[str] | None = None
    framework: str | None = None
    capabilities: list[Capability] | None = None
    domains: list[str] | None = None
    tags: list[str] | None = None
    created_at: datetime | None = None
    updated_at: datetime | None = None
    author: Author | None = None
    contributors: list[Contributor] | None = None
    links: list[Link] | None = None
    dependencies: list[Dependency] | None = None
    recommended_models: list[str] | None = None

    model_config = ConfigDict(extra="allow")


class CitationMetadata(BaseModel):
    """
    Represents an inline citation, providing info about information source. This
    is supposed to be rendered as an inline icon, optionally marking a text
    range it belongs to.

    If CitationMetadata is included together with content in the message part,
    the citation belongs to that content and renders at the MessagePart position.
    This way may be used for non-text content, like images and files.

    Alternatively, `start_index` and `end_index` may define a text range,
    counting characters in the current Message across all MessageParts with
    content type `text/*`, where the citation will be rendered. If one of
    `start_index` and `end_index` is missing or their values are equal, the
    citation renders only as an inline icon at that position.

    If both `start_index` and `end_index` are not present and MessagePart has no
    content, the citation renders as inline icon only at the MessagePart position.

    Properties:
    - url: URL of the source document.
    - title: Title of the source document.
    - description: Accompanying text, which may be a general description of the
                   source document, or a specific snippet.
    """

    kind: Literal["citation"] = "citation"
    start_index: Optional[int] = None
    end_index: Optional[int] = None
    url: Optional[str] = None
    title: Optional[str] = None
    description: Optional[str] = None


class TrajectoryMetadata(BaseModel):
    """
    Represents trajectory information for an agent's reasoning or tool execution
    steps. This metadata helps track the agent's decision-making process and
    provides transparency into how the agent arrived at its response.

    TrajectoryMetadata can capture either:
    1. A reasoning step with a message
    2. A tool execution with tool name, input, and output

    This information can be used for debugging, audit trails, and providing
    users with insight into the agent's thought process.

    Properties:
    - message: A reasoning step or thought in the agent's decision process.
    - tool_name: Name of the tool that was executed.
    - tool_input: Input parameters passed to the tool.
    - tool_output: Output or result returned by the tool.
    """

    kind: Literal["trajectory"] = "trajectory"
    message: Optional[str] = None
    tool_name: Optional[str] = None
    tool_input: Optional[AnyModel] = None
    tool_output: Optional[AnyModel] = None


class MessagePart(BaseModel):
    name: Optional[str] = None
    content_type: Optional[str] = "text/plain"
    content: Optional[str] = None
    content_encoding: Optional[Literal["plain", "base64"]] = "plain"
    content_url: Optional[AnyUrl] = None

    model_config = ConfigDict(extra="allow")

    metadata: Optional[CitationMetadata | TrajectoryMetadata] = Field(discriminator="kind", default=None)

    def model_post_init(self, __context: Any) -> None:
        if self.content is not None and self.content_url is not None:
            raise ValueError("Only one of content or content_url can be provided")


class Artifact(MessagePart):
    name: str


class Message(BaseModel):
    role: Literal["user"] | Literal["agent"] | str = Field("user", pattern=r"^(user|agent(\/[a-zA-Z0-9_\-]+)?)$")
    parts: list[MessagePart]
    created_at: datetime | None = Field(default_factory=lambda: datetime.now(timezone.utc))
    completed_at: datetime | None = Field(default_factory=lambda: datetime.now(timezone.utc))

    def __add__(self, other: "Message") -> "Message":
        if not isinstance(other, Message):
            raise TypeError(f"Cannot concatenate Message with {type(other).__name__}")
        if self.role != other.role:
            raise ValueError("Cannot concatenate messages with different roles")
        return Message(
            role=self.role,
            parts=self.parts + other.parts,
            created_at=min(self.created_at, other.created_at) if self.created_at and other.created_at else None,
            completed_at=max(self.completed_at, other.completed_at)
            if self.completed_at and other.completed_at
            else None,
        )

    def __str__(self) -> str:
        return "".join(
            part.content for part in self.parts if part.content is not None and part.content_type == "text/plain"
        )

    def compress(self) -> "Message":
        def can_be_joined(first: MessagePart, second: MessagePart) -> bool:
            return (
                first.name is None
                and second.name is None
                and first.content_type == "text/plain"
                and second.content_type == "text/plain"
                and first.content_encoding == "plain"
                and second.content_encoding == "plain"
                and first.content_url is None
                and second.content_url is None
            )

        def join(first: MessagePart, second: MessagePart) -> MessagePart:
            return MessagePart(
                name=None,
                content_type="text/plain",
                content=first.content + second.content,
                content_encoding="plain",
                content_url=None,
            )

        parts: list[MessagePart] = []
        for part in self.parts:
            if len(parts) > 0 and can_be_joined(parts[-1], part):
                parts[-1] = join(parts[-1], part)
            else:
                parts.append(part)
        return Message(parts=parts, created_at=self.created_at, completed_at=self.completed_at)


class RunMode(str, Enum):
    SYNC = "sync"
    ASYNC = "async"
    STREAM = "stream"


class RunStatus(str, Enum):
    CREATED = "created"
    IN_PROGRESS = "in-progress"
    AWAITING = "awaiting"
    CANCELLING = "cancelling"
    CANCELLED = "cancelled"
    COMPLETED = "completed"
    FAILED = "failed"

    @property
    def is_terminal(self) -> bool:
        terminal_states = {RunStatus.COMPLETED, RunStatus.FAILED, RunStatus.CANCELLED}
        return self in terminal_states


class MessageAwaitRequest(BaseModel):
    type: Literal["message"] = "message"
    message: Message


class MessageAwaitResume(BaseModel):
    type: Literal["message"] = "message"
    message: Message


AwaitRequest = Union[MessageAwaitRequest]
AwaitResume = Union[MessageAwaitResume]


class Run(BaseModel):
    run_id: RunId = Field(default_factory=uuid.uuid4)
    agent_name: AgentName
    session_id: SessionId | None = None
    status: RunStatus = RunStatus.CREATED
    await_request: AwaitRequest | None = None
    output: list[Message] = []
    error: Error | None = None
    created_at: datetime = Field(default_factory=lambda: datetime.now(timezone.utc))
    finished_at: datetime | None = None

    def raise_for_status(self) -> "Run":
        match self.status:
            case RunStatus.CANCELLED:
                raise asyncio.CancelledError()
            case RunStatus.FAILED:
                raise ACPError(error=self.error)
            case _:
                return self


class MessageCreatedEvent(BaseModel):
    type: Literal["message.created"] = "message.created"
    message: Message


class MessagePartEvent(BaseModel):
    type: Literal["message.part"] = "message.part"
    part: MessagePart


class ArtifactEvent(BaseModel):
    type: Literal["message.part"] = "message.part"
    part: Artifact


class MessageCompletedEvent(BaseModel):
    type: Literal["message.completed"] = "message.completed"
    message: Message


class RunAwaitingEvent(BaseModel):
    type: Literal["run.awaiting"] = "run.awaiting"
    run: Run


class GenericEvent(BaseModel):
    type: Literal["generic"] = "generic"
    generic: AnyModel


class RunCreatedEvent(BaseModel):
    type: Literal["run.created"] = "run.created"
    run: Run


class RunInProgressEvent(BaseModel):
    type: Literal["run.in-progress"] = "run.in-progress"
    run: Run


class RunFailedEvent(BaseModel):
    type: Literal["run.failed"] = "run.failed"
    run: Run


class RunCancelledEvent(BaseModel):
    type: Literal["run.cancelled"] = "run.cancelled"
    run: Run


class RunCompletedEvent(BaseModel):
    type: Literal["run.completed"] = "run.completed"
    run: Run


class ErrorEvent(BaseModel):
    type: Literal["error"] = "error"
    error: Error


Event = Union[
    ErrorEvent,
    RunCreatedEvent,
    RunInProgressEvent,
    MessageCreatedEvent,
    ArtifactEvent,
    MessagePartEvent,
    MessageCompletedEvent,
    RunAwaitingEvent,
    GenericEvent,
    RunCancelledEvent,
    RunFailedEvent,
    RunCompletedEvent,
]


class InputContentTypes(BaseModel):
    types: list[str]


class OutputContentTypes(BaseModel):
    types: list[str]


class AgentManifest(BaseModel):
    name: str
    description: str | None = None
    metadata: Metadata = Metadata()
    input_content_types: list[str] = Field(default_factory=lambda: ["*/*"])
    output_content_types: list[str] = Field(default_factory=lambda: ["*/*"])


class Session(BaseModel):
    id: SessionId = Field(default_factory=uuid.uuid4)
    history: list[ResourceUrl] = Field(default_factory=list)
    state: ResourceUrl | None = None

    loader: SkipJsonSchema[ResourceLoader | None] = Field(None, exclude=True)
    store: SkipJsonSchema[ResourceStore | None] = Field(None, exclude=True)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    async def load_history(self, *, loader: ResourceLoader | None = None) -> AsyncIterator[Message]:
        loader = loader or self.loader or ResourceLoader()
        for url in self.history:
            data = await loader.load(url)
            yield Message.model_validate_json(data)

    async def load_state(self, *, loader: ResourceLoader | None = None) -> bytes:
        loader = loader or self.loader or ResourceLoader()
        data = await loader.load(self.state)
        return data

    async def store_state(self, data: bytes, *, store: ResourceStore | None = None) -> ResourceUrl:
        store = store or self.store
        if not store:
            raise ValueError("Store must be specified")

        id = uuid.uuid4()
        await store.store(id, data)
        return await store.url(id)



# Copyright 2025 © BeeAI a Series of LF Projects, LLC
# SPDX-License-Identifier: Apache-2.0

from pydantic import BaseModel, ConfigDict


class AnyModel(BaseModel):
    model_config = ConfigDict(extra="allow")



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "ACP SDK"
copyright = "2025, IBM Corp."
author = "IBM Corp."

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["autoapi.extension"]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "alabaster"
html_static_path = ["_static"]

autoapi_dirs = ["../src/acp_sdk"]
autoapi_root = "api-reference"
autoapi_generate_api_docs = True



# Copyright 2025 © BeeAI a Series of LF Projects, LLC
# SPDX-License-Identifier: Apache-2.0

import asyncio
import base64
import os
import time
from collections.abc import AsyncGenerator, AsyncIterator, Generator
from datetime import timedelta
from threading import Thread

import pytest
import pytest_asyncio
import pytest_postgresql.factories
import pytest_redis.factories
from acp_sdk.models import Artifact, AwaitResume, Error, ErrorCode, Message, MessageAwaitRequest, MessagePart
from acp_sdk.models.errors import ACPError
from acp_sdk.server import Context, Server
from acp_sdk.server.store import MemoryStore, PostgreSQLStore, RedisStore, Store
from psycopg import AsyncConnection
from pytest_postgresql.executor import PostgreSQLExecutor
from pytest_postgresql.executor_noop import NoopExecutor
from pytest_redis.executor import NoopRedis, RedisExecutor
from redis.asyncio import Redis

from e2e.config import Config

REDIS_HOST = os.getenv("REDIS_HOST", None)
REDIS_PORT = os.getenv("REDIS_PORT", None)

POSTGRES_HOST = os.getenv("POSTGRES_HOST", None)
POSTGRES_PORT = os.getenv("POSTGRES_PORT", None)
POSTGRES_PASSWORD = os.getenv("POSTGRES_PASSWORD", None)

redis_db_proc = (
    pytest_redis.factories.redis_noproc(host=REDIS_HOST, port=int(REDIS_PORT))
    if REDIS_HOST and REDIS_PORT
    else pytest_redis.factories.redis_proc()
)

postgres_db_proc = (
    pytest_postgresql.factories.postgresql_noproc(
        host=POSTGRES_HOST, port=int(POSTGRES_PORT), password=POSTGRES_PASSWORD
    )
    if POSTGRES_HOST and POSTGRES_PORT and POSTGRES_PASSWORD
    else pytest_postgresql.factories.postgresql_proc()
)


@pytest_asyncio.fixture(scope="module", params=["memory", "redis", "postgres"])
async def store(
    request: pytest.FixtureRequest,
    redis_db_proc: RedisExecutor | NoopRedis,
    postgres_db_proc: PostgreSQLExecutor | NoopExecutor,
) -> AsyncGenerator[Store]:
    match request.param:
        case "memory":
            yield MemoryStore(limit=1000, ttl=timedelta(minutes=1))
        case "redis":
            redis = Redis(
                unix_socket_path=redis_db_proc.unixsocket,
            )
            yield RedisStore(redis=redis)
        case "postgres":
            aconn = await AsyncConnection.connect(
                f"user={postgres_db_proc.user} password={postgres_db_proc.password} host={postgres_db_proc.host} port={postgres_db_proc.port}"  # noqa: E501
            )
            async with aconn:
                yield PostgreSQLStore(aconn=aconn)
        case _:
            raise AssertionError()


@pytest.fixture(scope="module")
def server(request: pytest.FixtureRequest, store: Store) -> Generator[None]:
    server = Server()

    @server.agent()
    async def echo(input: list[Message], context: Context) -> AsyncIterator[Message]:
        for message in input:
            yield message

    @server.agent()
    async def slow_echo(input: list[Message], context: Context) -> AsyncIterator[Message]:
        for message in input:
            await asyncio.sleep(1)
            yield message

    @server.agent()
    async def history_echo(input: list[Message], context: Context) -> AsyncIterator[Message]:
        async for message in context.session.load_history():
            yield message
        for message in input:
            yield message

    @server.agent()
    async def awaiter(
        input: list[Message], context: Context
    ) -> AsyncGenerator[Message | MessageAwaitRequest, AwaitResume]:
        yield MessageAwaitRequest(message=Message(parts=[]))
        yield MessagePart(content="empty", content_type="text/plain")

    @server.agent()
    async def failer(input: list[Message], context: Context) -> AsyncIterator[Message]:
        yield Error(code=ErrorCode.INVALID_INPUT, message="Wrong question buddy!")
        raise RuntimeError("Unreachable code")

    @server.agent()
    async def raiser(input: list[Message], context: Context) -> AsyncIterator[Message]:
        raise ACPError(Error(code=ErrorCode.INVALID_INPUT, message="Wrong question buddy!"))

    @server.agent()
    async def sessioner(input: list[Message], context: Context) -> AsyncIterator[Message]:
        assert context.session is not None

        yield MessagePart(content=str(context.session.id), content_type="text/plain")

    @server.agent(
        input_content_types=["text/plain", "application/json"],
        output_content_types=["text/html", "application/json", "application/javascript", "text/css"],
    )
    async def mime_types(input: list[Message], context: Context) -> AsyncIterator[Message]:
        yield MessagePart(content="<h1>HTML Content</h1>", content_type="text/html")
        yield MessagePart(content='{"key": "value"}', content_type="application/json")
        yield MessagePart(content="console.log('Hello');", content_type="application/javascript")
        yield MessagePart(content="body { color: red; }", content_type="text/css")

    @server.agent()
    async def base64_encoding(input: list[Message], context: Context) -> AsyncIterator[Message]:
        yield Message(
            parts=[
                MessagePart(
                    content=base64.b64encode(
                        b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x01\x00\x00\x00\x01\x08\x06\x00\x00\x00\x1f\x15\xc4\x89\x00\x00\x00\nIDATx\x9cc\x00\x01\x00\x00\x05\x00\x01\r\n-\xb4\x00\x00\x00\x00IEND\xaeB`\x82"
                    ).decode("ascii"),
                    content_type="image/png",
                    content_encoding="base64",
                ),
                MessagePart(content="This is plain text", content_type="text/plain"),
            ]
        )

    @server.agent()
    async def artifact_producer(input: list[Message], context: Context) -> AsyncGenerator[Message | Artifact, None]:
        yield MessagePart(content="Processing with artifacts", content_type="text/plain")
        yield Artifact(name="text-result.txt", content_type="text/plain", content="This is a text artifact result")
        yield Artifact(
            name="data.json", content_type="application/json", content='{"results": [1, 2, 3], "status": "complete"}'
        )
        yield Artifact(
            name="image.png",
            content_type="image/png",
            content=base64.b64encode(
                b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x01\x00\x00\x00\x01\x08\x06\x00\x00\x00\x1f\x15\xc4\x89\x00\x00\x00\nIDATx\x9cc\x00\x01\x00\x00\x05\x00\x01\r\n-\xb4\x00\x00\x00\x00IEND\xaeB`\x82"
            ).decode("ascii"),
            content_encoding="base64",
        )

    thread = Thread(
        target=server.run, kwargs={"self_registration": False, "store": store, "port": Config.PORT}, daemon=True
    )
    thread.start()

    time.sleep(1)

    yield server

    server.should_exit = True
    thread.join(timeout=2)


@pytest.fixture(scope="module")
def multi_server(request: pytest.FixtureRequest) -> Generator[None]:
    server_one = Server()
    server_two = Server()

    @server_one.agent()
    @server_two.agent()
    async def history_echo(input: list[Message], context: Context) -> AsyncIterator[Message]:
        async for message in context.session.load_history():
            yield message
        for message in input:
            yield message

    thread_one = Thread(
        target=server_one.run, kwargs={"self_registration": False, "port": Config.PORT + 1}, daemon=True
    )
    thread_two = Thread(
        target=server_two.run, kwargs={"self_registration": False, "port": Config.PORT + 2}, daemon=True
    )
    thread_one.start()
    thread_two.start()

    time.sleep(1)

    yield (server_one, server_two)

    server_one.should_exit = True
    server_two.should_exit = True
    thread_one.join(timeout=2)
    thread_two.join(timeout=2)



# Copyright 2025 © BeeAI a Series of LF Projects, LLC
# SPDX-License-Identifier: Apache-2.0

import asyncio
import logging
import ssl
import typing
from collections.abc import AsyncIterator
from types import TracebackType
from typing import Self

import httpx
from httpx_sse import EventSource, aconnect_sse
from pydantic import TypeAdapter

from acp_sdk.client.types import Input
from acp_sdk.client.utils import input_to_messages
from acp_sdk.instrumentation import get_tracer
from acp_sdk.models import (
    ACPError,
    AgentManifest,
    AgentName,
    AgentReadResponse,
    AgentsListResponse,
    AwaitResume,
    Error,
    ErrorCode,
    ErrorEvent,
    Event,
    PingResponse,
    Run,
    RunCancelResponse,
    RunCreateRequest,
    RunCreateResponse,
    RunEventsListResponse,
    RunId,
    RunMode,
    RunResumeRequest,
    RunResumeResponse,
    Session,
    SessionReadResponse,
)

logger = logging.getLogger(__name__)


class Client:
    def __init__(
        self,
        *,
        session: Session | None = None,
        client: httpx.AsyncClient | None = None,
        manage_client: bool = True,
        auth: httpx._types.AuthTypes | None = None,
        params: httpx._types.QueryParamTypes | None = None,
        headers: httpx._types.HeaderTypes | None = None,
        cookies: httpx._types.CookieTypes | None = None,
        timeout: httpx._types.TimeoutTypes = None,
        verify: ssl.SSLContext | str | bool = True,
        cert: httpx._types.CertTypes | None = None,
        http1: bool = True,
        http2: bool = False,
        proxy: httpx._types.ProxyTypes | None = None,
        mounts: None | (typing.Mapping[str, httpx.AsyncBaseTransport | None]) = None,
        follow_redirects: bool = False,
        limits: httpx.Limits = httpx._config.DEFAULT_LIMITS,
        max_redirects: int = httpx._config.DEFAULT_MAX_REDIRECTS,
        event_hooks: None | (typing.Mapping[str, list[httpx._client.EventHook]]) = None,
        base_url: httpx.URL | str = "",
        transport: httpx.AsyncBaseTransport | None = None,
        trust_env: bool = True,
    ) -> None:
        self._session = session
        self._session_last_refresh_base_url: httpx.URL | None = None
        self._session_refresh_lock = asyncio.Lock()

        self._client = client or httpx.AsyncClient(
            auth=auth,
            params=params,
            headers=headers,
            cookies=cookies,
            timeout=timeout,
            verify=verify,
            cert=cert,
            http1=http1,
            http2=http2,
            proxy=proxy,
            mounts=mounts,
            follow_redirects=follow_redirects,
            limits=limits,
            max_redirects=max_redirects,
            event_hooks=event_hooks,
            base_url=base_url,
            transport=transport,
            trust_env=trust_env,
        )
        self._manage_client = manage_client

    @property
    def client(self) -> httpx.AsyncClient:
        return self._client

    async def __aenter__(self) -> Self:
        if self._manage_client:
            await self._client.__aenter__()
        self._session_span_manager = (
            (
                get_tracer()
                .start_as_current_span("session", attributes={"acp.session": str(self._session.id)})
                .__enter__()
            )
            if self._session
            else None
        )
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        if self._session_span_manager:
            self._session_span_manager.__exit__(exc_type, exc_value, traceback)
        if self._manage_client:
            await self._client.__aexit__(exc_type, exc_value, traceback)

    def session(self, session: Session | None = None) -> Self:
        return Client(client=self._client, manage_client=False, session=session or Session())

    async def agents(self, *, base_url: httpx.URL | str | None = None) -> AsyncIterator[AgentManifest]:
        response = await self._client.get(self._create_url("/agents", base_url=base_url))
        self._raise_error(response)
        for agent in AgentsListResponse.model_validate(response.json()).agents:
            yield agent

    async def agent(self, *, name: AgentName, base_url: httpx.URL | str | None = None) -> AgentManifest:
        response = await self._client.get(self._create_url(f"/agents/{name}", base_url=base_url))
        self._raise_error(response)
        response = AgentReadResponse.model_validate(response.json())
        return AgentManifest(**response.model_dump())

    async def ping(self, *, base_url: httpx.URL | str | None = None) -> bool:
        response = await self._client.get(self._create_url("/ping", base_url=base_url))
        self._raise_error(response)
        PingResponse.model_validate(response.json())
        return

    async def run_sync(self, input: Input, *, agent: AgentName, base_url: httpx.URL | str | None = None) -> Run:
        response = await self._client.post(
            self._create_url("/runs", base_url=base_url),
            content=RunCreateRequest(
                agent_name=agent,
                input=input_to_messages(input),
                mode=RunMode.SYNC,
                **(await self._prepare_session_for_run(base_url=base_url)),
            ).model_dump_json(),
        )
        self._raise_error(response)
        response = RunCreateResponse.model_validate(response.json())
        return Run(**response.model_dump())

    async def run_async(self, input: Input, *, agent: AgentName, base_url: httpx.URL | str | None = None) -> Run:
        response = await self._client.post(
            self._create_url("/runs", base_url=base_url),
            content=RunCreateRequest(
                agent_name=agent,
                input=input_to_messages(input),
                mode=RunMode.ASYNC,
                **(await self._prepare_session_for_run(base_url=base_url)),
            ).model_dump_json(),
        )
        self._raise_error(response)
        response = RunCreateResponse.model_validate(response.json())
        return Run(**response.model_dump())

    async def run_stream(
        self, input: Input, *, agent: AgentName, base_url: httpx.URL | str | None = None
    ) -> AsyncIterator[Event]:
        async with aconnect_sse(
            self._client,
            "POST",
            self._create_url("/runs", base_url=base_url),
            content=RunCreateRequest(
                agent_name=agent,
                input=input_to_messages(input),
                mode=RunMode.STREAM,
                **(await self._prepare_session_for_run(base_url=base_url)),
            ).model_dump_json(),
        ) as event_source:
            async for event in self._validate_stream(event_source):
                yield event

    async def run_status(self, *, run_id: RunId, base_url: httpx.URL | str | None = None) -> Run:
        response = await self._client.get(self._create_url(f"/runs/{run_id}", base_url=base_url))
        self._raise_error(response)
        return Run.model_validate(response.json())

    async def run_events(self, *, run_id: RunId, base_url: httpx.URL | str | None = None) -> AsyncIterator[Event]:
        response = await self._client.get(self._create_url(f"/runs/{run_id}/events", base_url=base_url))
        self._raise_error(response)
        response = RunEventsListResponse.model_validate(response.json())
        for event in response.events:
            yield event

    async def run_cancel(self, *, run_id: RunId, base_url: httpx.URL | str | None = None) -> Run:
        response = await self._client.post(self._create_url(f"/runs/{run_id}/cancel", base_url=base_url))
        self._raise_error(response)
        response = RunCancelResponse.model_validate(response.json())
        return Run(**response.model_dump())

    async def run_resume_sync(
        self, await_resume: AwaitResume, *, run_id: RunId, base_url: httpx.URL | str | None = None
    ) -> Run:
        response = await self._client.post(
            self._create_url(f"/runs/{run_id}", base_url=base_url),
            content=RunResumeRequest(await_resume=await_resume, mode=RunMode.SYNC).model_dump_json(),
        )
        self._raise_error(response)
        response = RunResumeResponse.model_validate(response.json())
        return Run(**response.model_dump())

    async def run_resume_async(
        self, await_resume: AwaitResume, *, run_id: RunId, base_url: httpx.URL | str | None = None
    ) -> Run:
        response = await self._client.post(
            self._create_url(f"/runs/{run_id}", base_url=base_url),
            content=RunResumeRequest(await_resume=await_resume, mode=RunMode.ASYNC).model_dump_json(),
        )
        self._raise_error(response)
        response = RunResumeResponse.model_validate(response.json())
        return Run(**response.model_dump())

    async def run_resume_stream(
        self, await_resume: AwaitResume, *, run_id: RunId, base_url: httpx.URL | str | None = None
    ) -> AsyncIterator[Event]:
        async with aconnect_sse(
            self._client,
            "POST",
            self._create_url(f"/runs/{run_id}", base_url=base_url),
            content=RunResumeRequest(await_resume=await_resume, mode=RunMode.STREAM).model_dump_json(),
        ) as event_source:
            async for event in self._validate_stream(event_source):
                yield event

    async def refresh_session(
        self, *, base_url: httpx.URL | str | None = None, timeout: httpx._types.TimeoutTypes = 5000
    ) -> Session:
        if not self._session:
            raise RuntimeError("Client is not in a session")

        async with self._session_refresh_lock:
            url = self._create_url(
                f"/sessions/{self._session.id}",
                base_url=base_url or self._session_last_refresh_base_url,
            )

            response = await self._client.get(url, timeout=timeout)
            self._raise_error(response)
            response = SessionReadResponse.model_validate(response.json())
            self._session = Session(**response.model_dump())
            return self._session

    async def _validate_stream(
        self,
        event_source: EventSource,
    ) -> AsyncIterator[Event]:
        if event_source.response.is_error:
            await event_source.response.aread()
            self._raise_error(event_source.response)
        async for event in event_source.aiter_sse():
            event: Event = TypeAdapter(Event).validate_json(event.data)
            if isinstance(event, ErrorEvent):
                raise ACPError(error=event.error)
            yield event

    def _raise_error(self, response: httpx.Response) -> None:
        try:
            response.raise_for_status()
        except httpx.HTTPError:
            raise ACPError(Error.model_validate(response.json()))

    def _create_base_url(self, base_url: httpx.URL | str | None) -> httpx.URL:
        base_url = httpx.URL(base_url or self._client.base_url)
        if not base_url.raw_path.endswith(b"/"):
            base_url = base_url.copy_with(raw_path=base_url.raw_path + b"/")
        return base_url

    def _create_url(self, endpoint: str, base_url: httpx.URL | str | None) -> httpx.URL:
        merge_url = httpx.URL(endpoint)

        if not merge_url.is_relative_url:
            raise ValueError("Endpoint must be a relative URL")

        base_url = self._create_base_url(base_url)
        merge_raw_path = base_url.raw_path + merge_url.raw_path.lstrip(b"/")
        return base_url.copy_with(raw_path=merge_raw_path)

    async def _prepare_session_for_run(self, *, base_url: httpx.URL | str | None) -> dict:
        if not self._session:
            return {}

        target_base_url = self._create_base_url(base_url=base_url)
        try:
            if self._session_last_refresh_base_url == target_base_url:
                # Same server, no need to forward session
                return {"session_id": self._session.id}

            session = await self.refresh_session(base_url=self._session_last_refresh_base_url or target_base_url)
            return {"session": session}
        except ACPError as e:
            if e.error.code == ErrorCode.NOT_FOUND:
                return {"session": self._session}
            raise e
        finally:
            await self._update_session_refresh_url(target_base_url)

    async def _update_session_refresh_url(self, url: httpx.URL) -> None:
        async with self._session_refresh_lock:
            self._session_last_refresh_base_url = url



# Copyright 2025 © BeeAI a Series of LF Projects, LLC
# SPDX-License-Identifier: Apache-2.0

import asyncio
from collections.abc import AsyncIterator
from typing import Generic

from psycopg import AsyncConnection
from psycopg.rows import dict_row

from acp_sdk.server.store.store import Store, StoreModel, T
from acp_sdk.server.store.utils import Stringable


class PostgreSQLStore(Store[T], Generic[T]):
    def __init__(self, *, aconn: AsyncConnection, table: str = "acp_store", channel: str = "acp_update") -> None:
        super().__init__()
        self._aconn = aconn
        self._table = table
        self._channel = channel

    async def get(self, key: Stringable) -> T | None:
        await self._ensure_table()
        async with self._aconn.cursor(row_factory=dict_row) as cur:
            await cur.execute(f"SELECT value FROM {self._table} WHERE key = %s", (str(key),))
            result = await cur.fetchone()
            if result is None:
                return None
            return StoreModel.model_validate(result["value"])

    async def set(self, key: Stringable, value: T | None) -> None:
        await self._ensure_table()
        async with self._aconn.cursor() as cur:
            if value is None:
                await cur.execute(
                    f"DELETE FROM {self._table} WHERE key = %s",
                    (str(key),),
                )
            else:
                await cur.execute(
                    f"""
                    INSERT INTO {self._table} (key, value)
                    VALUES (%s, %s)
                    ON CONFLICT (key)
                    DO UPDATE SET value = EXCLUDED.value
                    """,
                    (str(key), value.model_dump_json()),
                )
            await cur.execute(f"NOTIFY {self._channel}, '{key!s}'")  # NOTIFY appears not to accept params
            await self._aconn.commit()

    async def watch(self, key: Stringable, *, ready: asyncio.Event | None = None) -> AsyncIterator[T | None]:
        notify_conn = await AsyncConnection.connect(
            conninfo=f"{self._aconn.info.dsn} password={self._aconn.info.password}", autocommit=True
        )
        async with notify_conn:
            await notify_conn.execute(f"LISTEN {self._channel}")
            if ready:
                ready.set()
            async for notify in notify_conn.notifies():
                if notify.payload == str(key):
                    yield await self.get(key)

    async def _ensure_table(self) -> None:
        async with self._aconn.cursor() as cur:
            await cur.execute(f"""
                CREATE TABLE IF NOT EXISTS {self._table} (
                    key TEXT PRIMARY KEY,
                    value JSONB NOT NULL
                )
            """)
            await self._aconn.commit()



# Copyright 2025 © BeeAI a Series of LF Projects, LLC
# SPDX-License-Identifier: Apache-2.0

from acp_sdk.server.agent import AgentManifest as AgentManifest
from acp_sdk.server.agent import agent as agent
from acp_sdk.server.app import create_app as create_app
from acp_sdk.server.context import Context as Context
from acp_sdk.server.server import Server as Server
from acp_sdk.server.store import MemoryStore as MemoryStore
from acp_sdk.server.store import PostgreSQLStore as PostgreSQLStore
from acp_sdk.server.store import RedisStore as RedisStore
from acp_sdk.server.store import Store as Store
from acp_sdk.server.types import RunYield as RunYield
from acp_sdk.server.types import RunYieldResume as RunYieldResume



# Copyright 2025 © BeeAI a Series of LF Projects, LLC
# SPDX-License-Identifier: Apache-2.0


import pytest
from acp_sdk.shared.resources import ResourceLoader
from pytest_httpx import HTTPXMock

mock_resource = {"url": "http://invalid/resource1", "content": b"foobar"}


@pytest.mark.asyncio
async def test_resource_loader_load(httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=mock_resource["url"], method="GET", content=mock_resource["content"])

    resource_loader = ResourceLoader()
    resource = await resource_loader.load(mock_resource["url"])
    assert resource == mock_resource["content"]


@pytest.mark.asyncio
async def test_resource_loader_load_cache(httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=mock_resource["url"], method="GET", content=mock_resource["content"])

    resource_loader = ResourceLoader()
    resource = await resource_loader.load(mock_resource["url"])
    resource = await resource_loader.load(mock_resource["url"])

    assert resource == mock_resource["content"]



# Copyright 2025 © BeeAI a Series of LF Projects, LLC
# SPDX-License-Identifier: Apache-2.0

from acp_sdk.shared.resources import ResourceLoader as ResourceLoader
from acp_sdk.shared.resources import ResourceStore as ResourceStore



# Copyright 2025 © BeeAI a Series of LF Projects, LLC
# SPDX-License-Identifier: Apache-2.0

import asyncio
from collections.abc import AsyncIterator
from datetime import datetime
from typing import Generic

from cachetools import TTLCache

from acp_sdk.server.store.store import Store, StoreModel, T
from acp_sdk.server.store.utils import Stringable


class MemoryStore(Store[T], Generic[T]):
    def __init__(self, *, limit: int, ttl: int | None = None) -> None:
        super().__init__()
        self._cache: TTLCache[str, str] = TTLCache(maxsize=limit, ttl=ttl, timer=datetime.now)
        self._event = asyncio.Event()

    async def get(self, key: Stringable) -> T | None:
        value = self._cache.get(str(key))
        return StoreModel.model_validate_json(value) if value else value

    async def set(self, key: Stringable, value: T | None) -> None:
        if value is None:
            del self._cache[str(key)]
        else:
            self._cache[str(key)] = value.model_dump_json()
        self._event.set()

    async def watch(self, key: Stringable, *, ready: asyncio.Event | None = None) -> AsyncIterator[T | None]:
        if ready:
            ready.set()
        while True:
            await self._event.wait()
            self._event.clear()
            yield await self.get(key)



# Copyright 2025 © BeeAI a Series of LF Projects, LLC
# SPDX-License-Identifier: Apache-2.0

from datetime import timedelta

import httpx
import obstore
from async_lru import alru_cache
from obstore.store import AzureStore, GCSStore, HTTPStore, ObjectStore, S3Store

from acp_sdk.models.types import ResourceId, ResourceUrl


class ResourceLoader:
    def __init__(self, *, client: httpx.AsyncClient | None = None) -> None:
        self._client = client or httpx.AsyncClient(follow_redirects=False)

    @alru_cache()
    async def load(self, url: ResourceUrl) -> bytes:
        response = await self._client.get(str(url))
        response.raise_for_status()
        return await response.aread()


class ResourceStore:
    def __init__(self, *, store: ObjectStore, presigned_url_expiration: timedelta = timedelta(days=7)) -> None:
        self._store = store
        self._presigned_url_expiration = presigned_url_expiration

    async def load(self, id: ResourceId):  # noqa: ANN201
        result = await self._store.get_async(str(id))
        return result

    async def store(
        self,
        id: ResourceId,
        data: bytes,
    ) -> None:
        await self._store.put_async(str(id), data)

    async def url(self, id: ResourceId) -> ResourceUrl:
        if isinstance(self._store, (AzureStore, GCSStore, S3Store)):
            url = await obstore.sign_async(self._store, "GET", str(id), self._presigned_url_expiration)
            return ResourceUrl(url=url)
        elif isinstance(self._store, HTTPStore):
            return ResourceUrl(url=f"{self._store.url}/{id!s}")
        else:
            raise NotImplementedError("Unsupported store")



# Copyright 2025 © BeeAI a Series of LF Projects, LLC
# SPDX-License-Identifier: Apache-2.0

from acp_sdk.models import Message, MessagePart

Input = list[Message] | Message | list[MessagePart] | MessagePart | list[str] | str



# Copyright 2025 © BeeAI a Series of LF Projects, LLC
# SPDX-License-Identifier: Apache-2.0

import asyncio
from collections.abc import AsyncIterator
from typing import Generic

from redis.asyncio import Redis

from acp_sdk.server.store.store import Store, StoreModel, T
from acp_sdk.server.store.utils import Stringable


class RedisStore(Store[T], Generic[T]):
    def __init__(self, *, redis: Redis) -> None:
        super().__init__()
        self._redis = redis

    async def get(self, key: Stringable) -> T | None:
        value = await self._redis.get(str(key))
        return StoreModel.model_validate_json(value) if value else value

    async def set(self, key: Stringable, value: T | None) -> None:
        if value is None:
            await self._redis.delete(str(key))
        else:
            await self._redis.set(name=str(key), value=value.model_dump_json())

    async def watch(self, key: Stringable, *, ready: asyncio.Event | None = None) -> AsyncIterator[T]:
        await self._redis.config_set("notify-keyspace-events", "KEA")

        pubsub = self._redis.pubsub()
        channel = f"__keyspace@0__:{key!s}"
        await pubsub.subscribe(channel)
        if ready:
            ready.set()
        try:
            async for message in pubsub.listen():
                if message["type"] == "message":
                    yield await self.get(key)
        finally:
            await pubsub.unsubscribe(channel)
            await pubsub.aclose()



# Copyright 2025 © BeeAI a Series of LF Projects, LLC
# SPDX-License-Identifier: Apache-2.0

from typing import Any

from pydantic import BaseModel

from acp_sdk.models import AwaitRequest, AwaitResume, Message
from acp_sdk.models.models import MessagePart

RunYield = Message | MessagePart | str | AwaitRequest | BaseModel | dict[str | Any] | None | Exception
RunYieldResume = AwaitResume | None



# Copyright 2025 © BeeAI a Series of LF Projects, LLC
# SPDX-License-Identifier: Apache-2.0

from opentelemetry import trace

from acp_sdk.version import __version__


def get_tracer() -> trace.Tracer:
    return trace.get_tracer("acp-sdk", __version__)



# Copyright 2025 © BeeAI a Series of LF Projects, LLC
# SPDX-License-Identifier: Apache-2.0

from enum import Enum
from typing import Optional

from pydantic import BaseModel

from acp_sdk.models.common import AnyModel


class ErrorCode(str, Enum):
    SERVER_ERROR = "server_error"
    INVALID_INPUT = "invalid_input"
    NOT_FOUND = "not_found"


class Error(BaseModel):
    code: ErrorCode | str  # Allow arbitrary string for backwards compatibility
    message: str
    data: Optional[AnyModel] = None


class ACPError(Exception):
    def __init__(self, error: Error) -> None:
        super().__init__()
        self.error = error

    def __str__(self) -> str:
        return str(self.error.message)



# Copyright 2025 © BeeAI a Series of LF Projects, LLC
# SPDX-License-Identifier: Apache-2.0

from uuid import UUID

from pydantic import AnyHttpUrl

AgentName = str
SessionId = UUID
RunId = UUID
ResourceUrl = AnyHttpUrl
ResourceId = UUID



# Copyright 2025 © BeeAI a Series of LF Projects, LLC
# SPDX-License-Identifier: Apache-2.0

import pytest
from acp_sdk.client import Client
from acp_sdk.models import AgentManifest
from acp_sdk.server import Server


@pytest.mark.asyncio
async def test_ping(server: Server, client: Client) -> None:
    await client.ping()
    assert True


@pytest.mark.asyncio
async def test_agents_list(server: Server, client: Client) -> None:
    async for agent in client.agents():
        assert isinstance(agent, AgentManifest)


@pytest.mark.asyncio
async def test_agents_manifest(server: Server, client: Client) -> None:
    agent_name = "echo"
    agent = await client.agent(name=agent_name)
    assert isinstance(agent, AgentManifest)
    assert agent.name == agent_name


@pytest.mark.asyncio
async def test_input_content_types(server: Server, client: Client) -> None:
    agent_name = "mime_types"
    agent = await client.agent(name=agent_name)
    assert isinstance(agent, AgentManifest)
    assert agent.name == agent_name
    assert agent.input_content_types == ["text/plain", "application/json"]


@pytest.mark.asyncio
async def test_output_content_types(server: Server, client: Client) -> None:
    agent_name = "mime_types"
    agent = await client.agent(name=agent_name)
    assert isinstance(agent, AgentManifest)
    assert agent.name == agent_name
    assert agent.output_content_types == ["text/html", "application/json", "application/javascript", "text/css"]



# Copyright 2025 © BeeAI a Series of LF Projects, LLC
# SPDX-License-Identifier: Apache-2.0

import asyncio
import inspect
import logging
import uuid
from collections.abc import AsyncGenerator, AsyncIterator, Awaitable, Generator
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timezone
from typing import Callable, Self

import janus
from fastapi import Request
from pydantic import BaseModel, ValidationError

from acp_sdk.instrumentation import get_tracer
from acp_sdk.models import (
    ACPError,
    AnyModel,
    AwaitRequest,
    AwaitResume,
    Error,
    ErrorCode,
    Event,
    GenericEvent,
    Message,
    MessageCompletedEvent,
    MessageCreatedEvent,
    MessagePart,
    MessagePartEvent,
    ResourceId,
    ResourceUrl,
    Run,
    RunAwaitingEvent,
    RunCancelledEvent,
    RunCompletedEvent,
    RunCreatedEvent,
    RunFailedEvent,
    RunInProgressEvent,
    RunStatus,
    Session,
)
from acp_sdk.server.agent import AgentManifest
from acp_sdk.server.context import Context
from acp_sdk.server.logging import logger
from acp_sdk.server.store import Store
from acp_sdk.server.types import RunYield, RunYieldResume
from acp_sdk.shared import ResourceLoader, ResourceStore


class RunData(BaseModel):
    run: Run
    events: list[Event] = []

    @property
    def key(self) -> str:
        return str(self.run.run_id)

    async def watch(self, store: Store[Self], *, ready: asyncio.Event | None = None) -> AsyncIterator[Self]:
        async for data in store.watch(self.key, ready=ready):
            if data is None:
                raise RuntimeError("Missing data")
            yield data
            if data.run.status.is_terminal:
                break


class CancelData(BaseModel):
    pass


class Executor:
    def __init__(
        self,
        *,
        agent: AgentManifest,
        run_data: RunData,
        session: Session,
        executor: ThreadPoolExecutor,
        request: Request,
        run_store: Store[RunData],
        cancel_store: Store[CancelData],
        resume_store: Store[AwaitResume],
        session_store: Store[Session],
        resource_store: ResourceStore,
        resource_loader: ResourceLoader,
        create_resource_url: Callable[[ResourceId], Awaitable[ResourceUrl]],
    ) -> None:
        self.agent = agent
        self.session = session
        self.run_data = run_data
        self.executor = executor
        self.request = request

        self.run_store = run_store
        self.cancel_store = cancel_store
        self.resume_store = resume_store
        self.session_store = session_store
        self.resource_store = resource_store
        self.resource_loader = resource_loader

        self.create_resource_url = create_resource_url

        self.logger = logging.LoggerAdapter(logger, {"run_id": str(run_data.run.run_id)})

    def execute(self, input: list[Message], *, wait: asyncio.Event) -> None:
        self.task = asyncio.create_task(self._execute(input=input, executor=self.executor, wait=wait))
        self.watcher = asyncio.create_task(self._watch_for_cancellation())

    async def _push(self) -> None:
        await self.run_store.set(self.run_data.run.run_id, self.run_data)

    async def _emit(self, event: Event) -> None:
        freeze = event.model_copy(deep=True)
        self.run_data.events.append(freeze)
        await self._push()

    async def _await(self) -> AwaitResume:
        async for resume in self.resume_store.watch(self.run_data.key):
            if resume is not None:
                await self.resume_store.set(self.run_data.key, None)
                return resume

    async def _watch_for_cancellation(self) -> None:
        while not self.task.done():
            try:
                async for data in self.cancel_store.watch(self.run_data.key):
                    if data is not None:
                        self.task.cancel()
            except Exception:
                logger.warning("Cancellation watcher failed, restarting")

    async def _record_session(self, history: list[Message]) -> None:
        for message in history:
            id = uuid.uuid4()
            url = await self.create_resource_url(id)
            await self.resource_store.store(id, message.model_dump_json().encode())
            self.session.history.append(url)
        await self.session_store.set(self.session.id, self.session)

    async def _execute(self, input: list[Message], *, executor: ThreadPoolExecutor, wait: asyncio.Event) -> None:
        run_data = self.run_data
        with get_tracer().start_as_current_span("run"):
            in_message = False

            async def flush_message() -> None:
                nonlocal in_message
                if in_message:
                    message = run_data.run.output[-1]
                    message.completed_at = datetime.now(timezone.utc)
                    await self._emit(MessageCompletedEvent(message=message))
                    session_history.append(message)
                    in_message = False

            session_history = input.copy()
            try:
                await wait.wait()

                await self._emit(RunCreatedEvent(run=run_data.run))

                generator = self._execute_agent(
                    input=input,
                    session=self.session,
                    storage=self.resource_store,
                    loader=self.resource_loader,
                    executor=executor,
                    request=self.request,
                )
                self.logger.info("Run started")

                run_data.run.status = RunStatus.IN_PROGRESS
                await self._emit(RunInProgressEvent(run=run_data.run))

                await_resume = None
                while True:
                    next = await generator.asend(await_resume)

                    if isinstance(next, (MessagePart, str)):
                        if isinstance(next, str):
                            next = MessagePart(content=next)
                        if not in_message:
                            run_data.run.output.append(
                                Message(role=f"agent/{self.agent.name}", parts=[], completed_at=None)
                            )
                            in_message = True
                            await self._emit(MessageCreatedEvent(message=run_data.run.output[-1]))
                        run_data.run.output[-1].parts.append(next)
                        await self._emit(MessagePartEvent(part=next))
                    elif isinstance(next, Message):
                        await flush_message()
                        run_data.run.output.append(next.model_copy(update={"role": f"agent/{self.agent.name}"}))
                        await self._emit(MessageCreatedEvent(message=next))
                        for part in next.parts:
                            await self._emit(MessagePartEvent(part=part))
                        await self._emit(MessageCompletedEvent(message=next))
                        session_history.append(next)
                    elif isinstance(next, AwaitRequest):
                        run_data.run.await_request = next
                        run_data.run.status = RunStatus.AWAITING
                        await self._emit(RunAwaitingEvent(run=run_data.run))
                        self.logger.info("Run awaited")
                        await_resume = await self._await()
                        run_data.run.status = RunStatus.IN_PROGRESS
                        await self._emit(RunInProgressEvent(run=run_data.run))
                        self.logger.info("Run resumed")
                    elif isinstance(next, Error):
                        raise ACPError(error=next)
                    elif isinstance(next, BaseException):
                        raise next
                    elif next is None:
                        await flush_message()
                    elif isinstance(next, BaseModel):
                        await self._emit(GenericEvent(generic=AnyModel(**next.model_dump())))
                    else:
                        try:
                            generic = AnyModel.model_validate(next)
                            await self._emit(GenericEvent(generic=generic))
                        except ValidationError:
                            raise TypeError("Invalid yield")
            except StopAsyncIteration:
                await flush_message()
                run_data.run.status = RunStatus.COMPLETED
                run_data.run.finished_at = datetime.now(timezone.utc)
                try:
                    await self._record_session(session_history)
                except Exception as e:
                    self.logger.warning(f"Failed to record session: {e}")
                await self._emit(RunCompletedEvent(run=run_data.run))
                self.logger.info("Run completed")
            except asyncio.CancelledError:
                run_data.run.status = RunStatus.CANCELLED
                run_data.run.finished_at = datetime.now(timezone.utc)
                await self._emit(RunCancelledEvent(run=run_data.run))
                self.logger.info("Run cancelled")
            except Exception as e:
                if isinstance(e, ACPError):
                    run_data.run.error = e.error
                else:
                    run_data.run.error = Error(code=ErrorCode.SERVER_ERROR, message=str(e))
                run_data.run.status = RunStatus.FAILED
                run_data.run.finished_at = datetime.now(timezone.utc)
                await self._emit(RunFailedEvent(run=run_data.run))
                self.logger.exception("Run failed")

    async def _execute_agent(
        self,
        input: list[Message],
        session: Session,
        storage: ResourceStore,
        loader: ResourceLoader,
        executor: ThreadPoolExecutor,
        request: Request,
    ) -> AsyncGenerator[RunYield, RunYieldResume]:
        yield_queue: janus.Queue[RunYield] = janus.Queue()
        yield_resume_queue: janus.Queue[RunYieldResume] = janus.Queue()

        context = Context(
            session=session,
            store=storage,
            loader=loader,
            executor=executor,
            request=request,
            yield_queue=yield_queue,
            yield_resume_queue=yield_resume_queue,
        )

        if inspect.isasyncgenfunction(self.agent.run):
            run = asyncio.create_task(self._run_async_gen(input, context))
        elif inspect.iscoroutinefunction(self.agent.run):
            run = asyncio.create_task(self._run_coro(input, context))
        elif inspect.isgeneratorfunction(self.agent.run):
            run = asyncio.get_running_loop().run_in_executor(executor, self._run_gen, input, context)
        else:
            run = asyncio.get_running_loop().run_in_executor(executor, self._run_func, input, context)

        try:
            while not run.done() or yield_queue.async_q.qsize() > 0:
                value = yield await yield_queue.async_q.get()
                if isinstance(value, Exception):
                    raise value
                await yield_resume_queue.async_q.put(value)
        except janus.AsyncQueueShutDown:
            pass

    async def _run_async_gen(self, input: list[Message], context: Context) -> None:
        try:
            gen: AsyncGenerator[RunYield, RunYieldResume] = self.agent.run(input, context)
            value = None
            while True:
                value = await context.yield_async(await gen.asend(value))
        except StopAsyncIteration:
            pass
        except Exception as e:
            await context.yield_async(e)
        finally:
            context.shutdown()

    async def _run_coro(self, input: list[Message], context: Context) -> None:
        try:
            await context.yield_async(await self.agent.run(input, context))
        except Exception as e:
            await context.yield_async(e)
        finally:
            context.shutdown()

    def _run_gen(self, input: list[Message], context: Context) -> None:
        try:
            gen: Generator[RunYield, RunYieldResume] = self.agent.run(input, context)
            value = None
            while True:
                value = context.yield_sync(gen.send(value))
        except StopIteration:
            pass
        except Exception as e:
            context.yield_sync(e)
        finally:
            context.shutdown()

    def _run_func(self, input: list[Message], context: Context) -> None:
        try:
            context.yield_sync(self.agent.run(input, context))
        except Exception as e:
            context.yield_sync(e)
        finally:
            context.shutdown()



# Copyright 2025 © BeeAI a Series of LF Projects, LLC
# SPDX-License-Identifier: Apache-2.0

from acp_sdk.server.store.memory_store import MemoryStore as MemoryStore
from acp_sdk.server.store.postgresql_store import PostgreSQLStore as PostgreSQLStore
from acp_sdk.server.store.redis_store import RedisStore as RedisStore
from acp_sdk.server.store.store import Store as Store
